#include "llbmc.h"
#include <string>
#include <cstring>
#include <set>


// likeness_comparison
// -------------------

namespace likeness_comparison
{

inline bool like(const char* first, const char* second)
{
   return (!strncmp(first, second, strlen(second)));
}

inline bool like(std::string first, const char* second)
{
   return (!strncmp(first.c_str(), second, strlen(second)));
}

inline bool like(std::string first, std::string second)
{
   return (!strncmp(first.c_str(), second.c_str(), second.length()));
}

namespace private_implementation
{
   // C container
   // C::iterator i = container.find(...);
   // C::key_type = key(i);
   //
   template <class K, class V>
   const K& key(const std::pair<K,V>& elem)
   {
      return (elem.first);
   }

   template <class K>
   const K& key(const K& elem)
   {
      return elem;
   }

   // Case management
   class CaseSensitive
   {
   public:
      static int eq(char a, char b) { return a == b; }
      static int lt(char a, char b) { return a <  b; }
   };

   class NoCase // CaseInsensitive
   {
   public:
      static int eq(char a, char b) { return toupper(a) == toupper(b); }
      static int lt(char a, char b) { return toupper(a) <  toupper(b); }
   };

   template <class C> //  C = CaseSensitive or NoCase
   std::string::size_type first_different_char(const std::string& str1, const std::string& str2)
   {
      for (std::string::size_type i = 0; i < str1.length() && i < str2.length(); i++)
         if (!C::eq(str1[i],str2[i]))
            return i; //C::lt(str1[i],str2[i]) ? -1 : 1;

      return std::string::npos;
   } 

}

template <class C>   //requires Sorted_Associative_Container<C>
inline typename C::iterator find_like(C& container, std::string target)
{
   using namespace private_implementation;

   if (container.size() == 0)
      return container.end();

   typename C::iterator item_l = container.lower_bound(target);   
   
   if (item_l != container.end() && target == key(*item_l))
      return item_l;   // equal
   
   std::string::size_type char_pos;
   while(item_l != container.begin() &&
         (char_pos = first_different_char<CaseSensitive>(key(*(--item_l)),target)) != 0)
   {
      if (char_pos == std::string::npos)  // = like(target, key(*item_l))
         return item_l;
   }

   return container.end();
}


template <class C>   //requires Sorted_Associative_Container<C>
inline  bool has_like(C& container, std::string target)
{
   return (find_like(container, target) != container.end());
}


}


// -------------------------------------------------------
using namespace std;

int main()
{
    set<string> s;
    s.insert("A");
    s.insert("AA");
    s.insert("B");
	
    bool present = likeness_comparison::has_like(s, "AAA");

    assert(present == true);

    return 0;
}


